# Generated by Django 2.2.9 on 2020-02-09 16:38

import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import multiselectfield.db.fields
import structures.models


class Migration(migrations.Migration):

    dependencies = [
        ('structures', '0007_auto_20200123_1429'),
    ]

    operations = [
        migrations.CreateModel(
            name='EveCategory',
            fields=[
                ('id', models.IntegerField(help_text='Eve Online category ID', primary_key=True, serialize=False, validators=[django.core.validators.MinValueValidator(0)])),
                ('name', models.CharField(max_length=100)),
            ],
        ),
        migrations.AddField(
            model_name='owner',
            name='is_active',
            field=models.BooleanField(default=True, help_text='whether this owner is currently included in the sync process'),
        ),
        migrations.AddField(
            model_name='structure',
            name='eve_moon',
            field=models.ForeignKey(blank=True, default=None, help_text='Moon next to this structure - if any', null=True, on_delete=django.db.models.deletion.SET_DEFAULT, to='structures.EveMoon'),
        ),
        migrations.AddField(
            model_name='structure',
            name='eve_planet',
            field=models.ForeignKey(blank=True, default=None, help_text='Planet next to this structure - if any', null=True, on_delete=django.db.models.deletion.SET_DEFAULT, to='structures.EvePlanet'),
        ),
        migrations.AlterField(
            model_name='notification',
            name='notification_type',
            field=models.IntegerField(choices=[(401, 'MoonminingAutomaticFracture'), (402, 'MoonminingExtractionCancelled'), (403, 'MoonminingExtractionFinished'), (404, 'MoonminingExtractionStarted'), (405, 'MoonminingLaserFired'), (513, 'OwnershipTransferred'), (501, 'StructureAnchoring'), (502, 'StructureDestroyed'), (503, 'StructureFuelAlert'), (504, 'StructureLostArmor'), (505, 'StructureLostShields'), (506, 'StructureOnline'), (507, 'StructureServicesOffline'), (508, 'StructureUnanchoring'), (509, 'StructureUnderAttack'), (510, 'StructureWentHighPower'), (511, 'StructureWentLowPower'), (601, 'OrbitalAttacked'), (602, 'OrbitalReinforced'), (701, 'TowerAlertMsg'), (702, 'TowerResourceAlertMsg'), (801, 'EntosisCaptureStarted'), (802, 'SovCommandNodeEventStarted'), (803, 'SovAllClaimAquiredMsg'), (804, 'SovStructureReinforced'), (805, 'SovStructureDestroyed')]),
        ),
        migrations.AlterField(
            model_name='owner',
            name='is_alliance_main',
            field=models.BooleanField(default=False, help_text='whether alliance wide notifications are forwarded for this owner (e.g. sov notifications)'),
        ),
        migrations.AlterField(
            model_name='structure',
            name='state',
            field=models.IntegerField(blank=True, choices=[(1, 'anchor_vulnerable'), (2, 'anchoring'), (3, 'armor_reinforce'), (4, 'armor_vulnerable'), (5, 'deploy_vulnerable'), (6, 'fitting_invulnerable'), (7, 'hull_reinforce'), (8, 'hull_vulnerable'), (9, 'online_deprecated'), (10, 'onlining_vulnerable'), (11, 'shield_vulnerable'), (12, 'unanchored'), (21, 'offline'), (22, 'online'), (23, 'onlining'), (24, 'reinforced'), (25, 'unanchoring '), (0, 'N/A'), (13, 'unknown')], default=13, help_text='Current state of the structure'),
        ),
        migrations.AlterField(
            model_name='webhook',
            name='notification_types',
            field=multiselectfield.db.fields.MultiSelectField(choices=[(401, 'MoonminingAutomaticFracture'), (402, 'MoonminingExtractionCancelled'), (403, 'MoonminingExtractionFinished'), (404, 'MoonminingExtractionStarted'), (405, 'MoonminingLaserFired'), (513, 'OwnershipTransferred'), (501, 'StructureAnchoring'), (502, 'StructureDestroyed'), (503, 'StructureFuelAlert'), (504, 'StructureLostArmor'), (505, 'StructureLostShields'), (506, 'StructureOnline'), (507, 'StructureServicesOffline'), (508, 'StructureUnanchoring'), (509, 'StructureUnderAttack'), (510, 'StructureWentHighPower'), (511, 'StructureWentLowPower'), (601, 'OrbitalAttacked'), (602, 'OrbitalReinforced'), (701, 'TowerAlertMsg'), (702, 'TowerResourceAlertMsg'), (801, 'EntosisCaptureStarted'), (802, 'SovCommandNodeEventStarted'), (803, 'SovAllClaimAquiredMsg'), (804, 'SovStructureReinforced'), (805, 'SovStructureDestroyed')], default=structures.models.get_default_notification_types, help_text='only notifications which selected types are sent to this webhook', max_length=103),
        ),
        migrations.AddField(
            model_name='evegroup',
            name='eve_category',
            field=models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.SET_DEFAULT, to='structures.EveCategory'),
        ),
    ]




from .utils import clean_setting

# Whether to automatically add timers for certain notifications
# on the timerboard (will have no effect if aa-timerboard app is not installed)
STRUCTURES_ADD_TIMERS = \
    clean_setting('STRUCTURES_ADD_TIMERS', True)

# whether admins will get notifications about import events like
# when someone adds a structure owner
STRUCTURES_ADMIN_NOTIFICATIONS_ENABLED = \
    clean_setting('STRUCTURES_ADMIN_NOTIFICATIONS_ENABLED', True)

# Sets the default language to be used in case no language can be determined
# e.g. this language will be used when creating timers
# Please use the language codes as defined in the base.py settings file
STRUCTURES_DEFAULT_LANGUAGE = \
    clean_setting('STRUCTURES_DEFAULT_LANGUAGE', 'en')

# whether the structure list has default tags filter enabled by default
STRUCTURES_DEFAULT_TAGS_FILTER_ENABLED = \
    clean_setting('STRUCTURES_DEFAULT_TAGS_FILTER_ENABLED', False)

# Enables features for developers
# e.g. write access to all models in admin and writing raw data received from ESI
# UNDOCUMENTED SETTING
STRUCTURES_DEVELOPER_MODE = \
    clean_setting('STRUCTURES_DEVELOPER_MODE', False)

# Whether the customs offices feature is active
STRUCTURES_FEATURE_CUSTOMS_OFFICES = \
    clean_setting('STRUCTURES_FEATURE_CUSTOMS_OFFICES', True)

# Whether the starbases / POSes feature is active
STRUCTURES_FEATURE_STARBASES = \
    clean_setting('STRUCTURES_FEATURE_STARBASES', True)

# Max time in minutes since last successful notification forwarding
# before service is reported as down
STRUCTURES_FORWARDING_SYNC_GRACE_MINUTES = \
    clean_setting('STRUCTURES_FORWARDING_SYNC_GRACE_MINUTES', 5)

# defines after how many hours a notification becomes stale
# stale notification will no longer be sent automatically
STRUCTURES_HOURS_UNTIL_STALE_NOTIFICATION = \
    clean_setting('STRUCTURES_HOURS_UNTIL_STALE_NOTIFICATION', 24)

# whether to create / remove timers from moon extraction notifications
STRUCTURES_MOON_EXTRACTION_TIMERS_ENABLED = \
    clean_setting('STRUCTURES_MOON_EXTRACTION_TIMERS_ENABLED', True)

# Max number of retries for sending a notification if an error occurred
# e.g. rate limiting
STRUCTURES_NOTIFICATION_MAX_RETRIES = \
    clean_setting('STRUCTURES_NOTIFICATION_MAX_RETRIES', 3)

# Max time in minutes since last successful notification sync
# before service is reported as down
STRUCTURES_NOTIFICATION_SYNC_GRACE_MINUTES = \
    clean_setting('STRUCTURES_NOTIFICATION_SYNC_GRACE_MINUTES', 15)

# Default wait time in seconds before retrying to send a notification
# to Discord after an error occurred
STRUCTURES_NOTIFICATION_WAIT_SEC = \
    clean_setting('STRUCTURES_NOTIFICATION_WAIT_SEC', 5)

# Enables archiving of all notifications received from ESI to files
# notifications will by stored into one continuous file per corporations
# UNDOCUMENTED SETTING
STRUCTURES_NOTIFICATIONS_ARCHIVING_ENABLED = \
    clean_setting('STRUCTURES_NOTIFICATIONS_ARCHIVING_ENABLED', False)

# how to handle notification about NPC attacks
STRUCTURES_REPORT_NPC_ATTACKS = \
    clean_setting('STRUCTURES_REPORT_NPC_ATTACKS', True)

# whether fuel expires in structures browser is shown as absolute value
STRUCTURES_SHOW_FUEL_EXPIRES_RELATIVE = \
    clean_setting('STRUCTURES_SHOW_FUEL_EXPIRES_RELATIVE', True)

# Max time in minutes since last successful structures sync
# before service is reported as down
STRUCTURES_STRUCTURE_SYNC_GRACE_MINUTES = \
    clean_setting('STRUCTURES_STRUCTURE_SYNC_GRACE_MINUTES', 120)

# whether created timers are corp restricted on the timerboard
STRUCTURES_TIMERS_ARE_CORP_RESTRICTED = \
    clean_setting('STRUCTURES_TIMERS_ARE_CORP_RESTRICTED', False)



import socket
from datetime import timedelta
import logging
import os

from django.apps import apps
from django.conf import settings
from django.contrib.auth.models import User, Permission
from django.contrib.messages.constants  \
    import DEBUG, ERROR, SUCCESS, WARNING, INFO
from django.contrib import messages
from django.db.models import Q
from django.test import TestCase
from django.utils.functional import lazy
from django.utils.html import format_html, mark_safe
from django.utils.translation import gettext_lazy as _

from allianceauth.notifications import notify


# Format for output of datetime for this app
DATETIME_FORMAT = '%Y-%m-%d %H:%M'

format_html_lazy = lazy(format_html, str)


class LoggerAddTag(logging.LoggerAdapter):
    """add custom tag to a logger"""
    def __init__(self, logger, prefix):
        super(LoggerAddTag, self).__init__(logger, {})
        self.prefix = prefix

    def process(self, msg, kwargs):
        return '[%s] %s' % (self.prefix, msg), kwargs


logger = LoggerAddTag(logging.getLogger(__name__), __package__)


def get_swagger_spec_path() -> str:
    """returns the path to the current swagger spec file"""
    return os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        'swagger.json'
    )


def make_logger_prefix(tag: str):
    """creates a function to add logger prefix, which returns tag when used empty"""
    return lambda text = '': '{}{}'.format(tag, (': ' + text) if text else '')


class messages_plus():
    """Pendant to Django messages adding level icons and HTML support

    Careful: Use with safe strings only
    """
    _glyph_map = {
        DEBUG: 'eye-open',
        INFO: 'info-sign',
        SUCCESS: 'ok-sign',
        WARNING: 'exclamation-sign',
        ERROR: 'alert',
    }

    @classmethod
    def _add_messages_icon(cls, level: int, message: str) -> str:
        """Adds an level based icon to standard Django messages"""
        if level not in cls._glyph_map:
            raise ValueError('glyph for level not defined')
        else:
            glyph = cls._glyph_map[level]

        return format_html(
            '<span class="glyphicon glyphicon-{}" '
            'aria-hidden="true"></span>&nbsp;&nbsp;{}',
            glyph,
            message
        )

    @classmethod
    def debug(
        cls,
        request: object,
        message: str,
        extra_tags: str = '',
        fail_silently: bool = False
    ):
        messages.debug(
            request,
            cls._add_messages_icon(DEBUG, message),
            extra_tags,
            fail_silently
        )

    @classmethod
    def info(
        cls,
        request: object,
        message: str,
        extra_tags: str = '',
        fail_silently: bool = False
    ):
        messages.info(
            request,
            cls._add_messages_icon(INFO, message),
            extra_tags,
            fail_silently
        )

    @classmethod
    def success(
        cls,
        request: object,
        message: str,
        extra_tags: str = '',
        fail_silently: bool = False
    ):
        messages.success(
            request,
            cls._add_messages_icon(SUCCESS, message),
            extra_tags,
            fail_silently
        )

    @classmethod
    def warning(
        cls,
        request: object,
        message: str,
        extra_tags: str = '',
        fail_silently: bool = False
    ):
        messages.warning(
            request,
            cls._add_messages_icon(WARNING, message),
            extra_tags,
            fail_silently
        )

    @classmethod
    def error(
        cls,
        request: object,
        message: str,
        extra_tags: str = '',
        fail_silently: bool = False
    ):
        messages.error(
            request,
            cls._add_messages_icon(ERROR, message),
            extra_tags,
            fail_silently
        )


def notify_admins(message: str, title: str, level='info') -> None:
    """send notification to all admins"""
    try:
        perm = Permission.objects.get(codename="logging_notifications")
        users = User.objects\
            .filter(
                Q(groups__permissions=perm)
                | Q(user_permissions=perm)
                | Q(is_superuser=True)
            ).distinct()

        for user in users:
            notify(
                user,
                title=title,
                message=message,
                level=level
            )
    except Permission.DoesNotExist:
        pass


def chunks(lst, size):
    """Yield successive sized chunks from lst."""
    for i in range(0, len(lst), size):
        yield lst[i:i + size]


def clean_setting(
    name: str,
    default_value: object,
    min_value: int = None,
    max_value: int = None,
    required_type: type = None
):
    """cleans the input for a custom setting

    Will use `default_value` if settings does not exit or has the wrong type
    or is outside define boundaries (for int only)

    Need to define `required_type` if `default_value` is `None`

    Will assume `min_value` of 0 for int (can be overriden)

    Returns cleaned value for setting
    """
    if default_value is None and not required_type:
        raise ValueError('You must specify a required_type for None defaults')

    if not required_type:
        required_type = type(default_value)

    if min_value is None and required_type == int:
        min_value = 0

    if not hasattr(settings, name):
        cleaned_value = default_value
    else:
        if (isinstance(getattr(settings, name), required_type)
            and (min_value is None or getattr(settings, name) >= min_value)
            and (max_value is None or getattr(settings, name) <= max_value)
        ):
            cleaned_value = getattr(settings, name)
        else:
            logger.warn(
                'You setting for {} it not valid. Please correct it. '
                'Using default for now: {}'.format(
                    name,
                    default_value
                )
            )
            cleaned_value = default_value
    return cleaned_value


def set_test_logger(logger_name: str, name: str) -> object:
    """set logger for current test module
    
    Args:
    - logger: current logger object
    - name: name of current module, e.g. __file__
    
    Returns:
    - amended logger
    """
    
    # reconfigure logger so we get logging from tested module
    f_format = logging.Formatter(
        '%(asctime)s - %(levelname)s - %(module)s:%(funcName)s - %(message)s'
    )
    f_handler = logging.FileHandler(
        '{}.log'.format(os.path.splitext(name)[0]),
        'w+'
    )
    f_handler.setFormatter(f_format)
    logger = logging.getLogger(logger_name)
    logger.level = logging.DEBUG
    logger.addHandler(f_handler)
    logger.propagate = False
    return logger


def timeuntil_str(duration: timedelta) -> str:
    """return the duration as nicely formatted string. 
    Or empty string if duration is negative. 
    
    Format: '[[[999y] [99m]] 99d] 99h 99m 99s'
    """
    seconds = int(duration.total_seconds())
    if seconds > 0:
        periods = [
            # Translators: Abbreviation for years
            (_('y'), 60 * 60 * 24 * 365, False),
            
            # Translators: Abbreviation for months
            (_('mt'), 60 * 60 * 24 * 30, False),
            
            # Translators: Abbreviation for days
            (_('d'), 60 * 60 * 24, False),
            
            # Translators: Abbreviation for hours
            (_('h'), 60 * 60, True),
            
            # Translators: Abbreviation for months
            (_('m'), 60, True),
            
            # Translators: Abbreviation for seconds
            (_('s'), 1, True)
        ]
        strings = list()
        for period_name, period_seconds, period_static in periods:
            if seconds >= period_seconds or period_static:
                period_value, seconds = divmod(seconds, period_seconds)            
                strings.append('{}{}'.format(period_value, period_name))

        result = ' '.join(strings)
    else:
        result = ''

    return result


class SocketAccessError(Exception):
    pass


class NoSocketsTestCase(TestCase):
    """Variation of TestCase class that prevents any use of sockets"""
    
    @classmethod
    def setUpClass(cls):
        cls.socket_original = socket.socket
        socket.socket = cls.guard
        return super().setUpClass()

    @classmethod
    def tearDownClass(cls):
        socket.socket = cls.socket_original
        return super().tearDownClass()

    @staticmethod
    def guard(*args, **kwargs):
        raise SocketAccessError('Attempted to access network')


def app_labels() -> set:
    """returns set of all current app labels"""
    return {x for x in apps.app_configs.keys()}


def add_no_wrap_html(text: str) -> str:
    """add no-wrap HTML to text"""
    return format_html(
        '<span style="white-space: nowrap;">{}</span>', mark_safe(text)
    )


def create_bs_label_html(text: str, style: str) -> str:
    """create Bootstrap label and return HTML"""
    return format_html(
        '<span class="label label-{}">{}</span>', style, text
    )


def yesno_str(value: bool) -> str:
    """returns yes/no for boolean else empty string and with localization"""
    if value is True:
        return _('yes') 
    elif value is False:
        return _('no')
    else:
        return ''



# Generated by Django 2.2.8 on 2020-01-23 14:29

import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import multiselectfield.db.fields
import structures.models


class Migration(migrations.Migration):

    dependencies = [
        ('structures', '0006_auto_20200120_0147'),
    ]

    operations = [
        migrations.AlterField(
            model_name='notification',
            name='notification_type',
            field=models.IntegerField(choices=[(401, 'MoonminingAutomaticFracture'), (402, 'MoonminingExtractionCancelled'), (403, 'MoonminingExtractionFinished'), (404, 'MoonminingExtractionStarted'), (405, 'MoonminingLaserFired'), (513, 'OwnershipTransferred'), (501, 'StructureAnchoring'), (502, 'StructureDestroyed'), (503, 'StructureFuelAlert'), (504, 'StructureLostArmor'), (505, 'StructureLostShields'), (506, 'StructureOnline'), (507, 'StructureServicesOffline'), (508, 'StructureUnanchoring'), (509, 'StructureUnderAttack'), (510, 'StructureWentHighPower'), (511, 'StructureWentLowPower'), (601, 'OrbitalAttacked'), (602, 'OrbitalReinforced')]),
        ),
        migrations.AlterField(
            model_name='webhook',
            name='notification_types',
            field=multiselectfield.db.fields.MultiSelectField(choices=[(401, 'MoonminingAutomaticFracture'), (402, 'MoonminingExtractionCancelled'), (403, 'MoonminingExtractionFinished'), (404, 'MoonminingExtractionStarted'), (405, 'MoonminingLaserFired'), (513, 'OwnershipTransferred'), (501, 'StructureAnchoring'), (502, 'StructureDestroyed'), (503, 'StructureFuelAlert'), (504, 'StructureLostArmor'), (505, 'StructureLostShields'), (506, 'StructureOnline'), (507, 'StructureServicesOffline'), (508, 'StructureUnanchoring'), (509, 'StructureUnderAttack'), (510, 'StructureWentHighPower'), (511, 'StructureWentLowPower'), (601, 'OrbitalAttacked'), (602, 'OrbitalReinforced')], default=structures.models.get_default_notification_types, help_text='only notifications which selected types are sent to this webhook', max_length=75),
        ),
        migrations.CreateModel(
            name='EvePlanet',
            fields=[
                ('id', models.IntegerField(help_text='Eve Online item ID', primary_key=True, serialize=False, validators=[django.core.validators.MinValueValidator(0)])),
                ('name', models.CharField(max_length=100)),
                ('position_x', models.FloatField(blank=True, default=None, help_text='x position of the structure in the solar system', null=True)),
                ('position_y', models.FloatField(blank=True, default=None, help_text='y position of the structure in the solar system', null=True)),
                ('position_z', models.FloatField(blank=True, default=None, help_text='z position of the structure in the solar system', null=True)),
                ('eve_solar_system', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='structures.EveSolarSystem')),
                ('eve_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='structures.EveType')),
            ],
        ),
    ]



# Generated by Django 2.2.5 on 2019-11-29 03:08

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('structures', '0003_auto_20191125_2157'),
    ]

    operations = [
        migrations.AlterField(
            model_name='structure',
            name='reinforce_hour',
            field=models.IntegerField(blank=True, default=None, help_text='The hour of day that determines the four hour window when the structure will randomly exit its reinforcement periods and become vulnerable to attack against its armor and/or hull. The structure will become vulnerable at a random time that is +/- 2 hours centered on the value of this property', null=True, validators=[django.core.validators.MinValueValidator(0), django.core.validators.MaxValueValidator(23)]),
        ),
        migrations.AlterField(
            model_name='structure',
            name='state',
            field=models.IntegerField(blank=True, choices=[(0, 'N/A'), (1, 'anchor_vulnerable'), (2, 'anchoring'), (3, 'armor_reinforce'), (4, 'armor_vulnerable'), (5, 'deploy_vulnerable'), (6, 'fitting_invulnerable'), (7, 'hull_reinforce'), (8, 'hull_vulnerable'), (9, 'online_deprecated'), (10, 'onlining_vulnerable'), (11, 'shield_vulnerable'), (12, 'unanchored'), (13, 'unknown')], default=13, help_text='Current state of the structure'),
        ),
    ]



from celery import chain

from django.contrib import admin
from django.db import models
from django.db.models.functions import Lower
from django.utils.html import format_html

from allianceauth.eveonline.models import EveCorporationInfo, EveAllianceInfo
from allianceauth.services.hooks import get_extension_logger

from . import __title__
from .app_settings import STRUCTURES_DEVELOPER_MODE
from .models import (
    EveCategory,
    EveGroup,
    EveType,
    EveRegion,
    EveConstellation,
    EveSolarSystem,
    EveMoon,
    EvePlanet,
    EveSovereigntyMap,
    StructureTag,
    StructureService,
    Webhook,
    EveEntity,
    Owner,
    Notification,
    Structure
)
from . import tasks
from .utils import LoggerAddTag


logger = LoggerAddTag(get_extension_logger(__name__), __title__)


if STRUCTURES_DEVELOPER_MODE:
    @admin.register(EveConstellation)
    class EveConstellationAdmin(admin.ModelAdmin):
        pass

    @admin.register(EveEntity)
    class EveEntityAdmin(admin.ModelAdmin):
        list_display = (
            'id',
            'name',
            'category',
        )
        list_filter = ('category',)
        list_display_links = None

    @admin.register(EveCategory)
    class EveCategoryAdmin(admin.ModelAdmin):
        pass

    @admin.register(EveGroup)
    class EveGroupAdmin(admin.ModelAdmin):
        pass

    @admin.register(EveMoon)
    class EveMoonAdmin(admin.ModelAdmin):
        pass

    @admin.register(EvePlanet)
    class EvePlanetAdmin(admin.ModelAdmin):
        pass

    @admin.register(EveRegion)
    class EveRegionAdmin(admin.ModelAdmin):
        pass

    @admin.register(EveSolarSystem)
    class EveSolarSystemAdmin(admin.ModelAdmin):
        pass

    @admin.register(EveType)
    class EveTypeAdmin(admin.ModelAdmin):
        pass

    @admin.register(EveSovereigntyMap)
    class EveSovereigntyMapAdmin(admin.ModelAdmin):
        list_display = (
            'solar_system_id', 'alliance_id', 'corporation_id', 'faction_id'
        )
        search_fields = [
            'solar_system_id', 'alliance_id', 'corporation_id', 'faction_id'
        ]


@admin.register(Notification)
class NotificationAdmin(admin.ModelAdmin):
    list_display = (
        'notification_id',
        'owner',
        'notification_type',
        'timestamp',
        'created',
        'last_updated',
        '_webhooks',
        'is_sent',
        'is_timer_added'
    )
    ordering = ['-timestamp', '-notification_id']
    list_filter = (
        'owner',
        'notification_type',
        'is_sent'
    )
    
    def _webhooks(self, obj):
        names = [x.name for x in obj.owner.webhooks.all().order_by('name')]
        if names:
            return ', '.join(names)
        else:
            return None

    _webhooks.short_description = 'Webhooks'

    actions = (
        'mark_as_sent',
        'mark_as_unsent',
        'send_to_webhook',
        'process_for_timerboard'
    )

    def mark_as_sent(self, request, queryset):
        notifications_count = 0
        for obj in queryset:
            obj.is_sent = True
            obj.save()
            notifications_count += 1

        self.message_user(
            request,
            '{} notifications marked as sent'.format(notifications_count)
        )

    mark_as_sent.short_description = "Mark selected notifications as sent"

    def mark_as_unsent(self, request, queryset):
        notifications_count = 0
        for obj in queryset:
            obj.is_sent = False
            obj.save()
            notifications_count += 1

        self.message_user(
            request,
            '{} notifications marked as unsent'.format(notifications_count)
        )

    mark_as_unsent.short_description = \
        "Mark selected notifications as unsent"

    def send_to_webhook(self, request, queryset):
        obj_pks = [obj.pk for obj in queryset]        
        tasks.send_notifications.delay(obj_pks)            

        self.message_user(
            request,
            'Initiated sending of {} notifications to '
            'configured webhooks'.format(len(obj_pks))
        )

    send_to_webhook.short_description = \
        "Send selected notifications to configured webhooks"

    def process_for_timerboard(self, request, queryset):
        notifications_count = 0
        for obj in queryset:
            if obj.process_for_timerboard():
                notifications_count += 1

        self.message_user(
            request,
            'Added timers from {} notifications to timerboard'.format(
                notifications_count
            )
        )

    process_for_timerboard.short_description = \
        "Process selected notifications for timerboard"

    def has_add_permission(self, request):
        return True if STRUCTURES_DEVELOPER_MODE else False

    def has_change_permission(self, request, obj=None):
        return True if STRUCTURES_DEVELOPER_MODE else False


class OwnerSyncStatusFilter(admin.SimpleListFilter):
    title = 'is sync ok'

    parameter_name = 'sync_status__exact'

    def lookups(self, request, model_admin):
        """List of values to allow admin to select"""
        return (
            ('yes', 'Yes'),
            ('no', 'No'),
        )

    def queryset(self, request, queryset):
        """Return the filtered queryset"""
        if self.value() == 'yes':
            return queryset.filter(
                structures_last_error=Owner.ERROR_NONE,
                notifications_last_error=Owner.ERROR_NONE,
                forwarding_last_error=Owner.ERROR_NONE
            )
        elif self.value() == 'no':
            return queryset.exclude(
                structures_last_error=Owner.ERROR_NONE,
                notifications_last_error=Owner.ERROR_NONE,
                forwarding_last_error=Owner.ERROR_NONE
            )
        else:
            return queryset


@admin.register(Owner)
class OwnerAdmin(admin.ModelAdmin):
    list_select_related = True
    list_display = (        
        '_corporation',        
        '_alliance',
        'character',
        '_webhooks',
        '_is_active',
        '_is_structure_sync_ok',
        '_is_notification_sync_ok',
        '_is_forwarding_sync_ok',
        
    )
    list_filter = (
        ('corporation__alliance', admin.RelatedOnlyFieldListFilter),
        'is_active',
        OwnerSyncStatusFilter,
    )
    ordering = ['corporation__corporation_name']
    fieldsets = (
        (None, {
            'fields': (
                'corporation',
                'character',
                'webhooks',
                'is_alliance_main',
                'is_included_in_service_status',
                'is_active',
            )
        }),
        ('Sync Status', {
            'classes': ('collapse',),
            'fields': (
                ('structures_last_sync', 'structures_last_error', ),
                ('notifications_last_sync', 'notifications_last_error', ),
                ('forwarding_last_sync', 'forwarding_last_error', ),
            )
        }),
    )

    def _corporation(self, obj):        
        return obj.corporation.corporation_name
    
    _corporation.admin_order_field = 'corporation__corporation_name'
        
    def _alliance(self, obj):
        if obj.corporation.alliance:
            return obj.corporation.alliance.alliance_name
        else:
            return None

    _alliance.admin_order_field = 'corporation__alliance__alliance_name'

    def _webhooks(self, obj):
        webhook_names = [x.name for x in obj.webhooks.all().order_by('name')]
        if webhook_names:
            return ', '.join(webhook_names)
        else:
            return None

    _webhooks.short_description = 'Webhooks'

    def _is_active(self, obj):
        return obj.is_active
    
    _is_active.boolean = True
    _is_active.short_description = 'active'

    def _is_structure_sync_ok(self, obj):
        if not obj.is_active:
            return None
        else:
            return obj.is_structure_sync_ok

    _is_structure_sync_ok.boolean = True
    _is_structure_sync_ok.short_description = 'structure sync'

    def _is_notification_sync_ok(self, obj):
        if not obj.is_active:
            return None
        else:
            return obj.is_notification_sync_ok

    _is_notification_sync_ok.boolean = True
    _is_notification_sync_ok.short_description = 'notification sync'

    def _is_forwarding_sync_ok(self, obj):
        if not obj.is_active:
            return None
        else:
            return obj.is_forwarding_sync_ok

    _is_forwarding_sync_ok.boolean = True
    _is_forwarding_sync_ok.short_description = 'forwarding'

    def get_readonly_fields(self, request, obj=None):
        if obj:    # editing an existing object
            return self.readonly_fields + (
                'notifications_last_error',
                'notifications_last_sync',
                'structures_last_error',
                'structures_last_sync',
                'forwarding_last_sync',
                'forwarding_last_error',
            )
        return self.readonly_fields

    actions = (
        'update_structures',
        'fetch_notifications',
        'send_notifications'
    )

    def update_structures(self, request, queryset):
        for obj in queryset:
            tasks.update_structures_for_owner.delay(
                obj.pk, user_pk=request.user.pk
            )
            text = 'Started updating structures for: {}. '.format(obj)
            text += 'You will receive a notification once it is completed.'

            self.message_user(
                request,
                text
            )

    update_structures.short_description = "Update structures from EVE server"

    def fetch_notifications(self, request, queryset):
        for obj in queryset:
            tasks.fetch_notifications_for_owner.delay(
                obj.pk, user_pk=request.user.pk
            )
            text = 'Started fetching notifications for: {}. '.format(obj)
            text += 'You will receive a notification once it is completed.'

            self.message_user(
                request,
                text
            )

    fetch_notifications.short_description = \
        "Fetch notifications from EVE server"

    def send_notifications(self, request, queryset):
        send_tasks = list()
        for owner in queryset:            
            send_tasks.append(tasks.send_new_notifications_for_owner.si(
                owner_pk=owner.pk
            ))
            self.message_user(
                request,
                'Started sending new notifications for: {}. '.format(owner)
            )
        chain(send_tasks).delay()

    send_notifications.short_description = "Send new notifications to Discord"

    def has_add_permission(self, request):
        return True if STRUCTURES_DEVELOPER_MODE else False


@admin.register(StructureTag)
class StructureTagAdmin(admin.ModelAdmin):
    list_display = (
        'name',
        'description',        
        'order',        
        'style',
        'is_default',
        'is_user_managed',
    )
    list_filter = (        
        'is_default',        
        'style',
        'is_user_managed',
    )    
    readonly_fields = ('is_user_managed',)

    def has_delete_permission(self, request, obj=None):
        return False if obj and not obj.is_user_managed else True

    def has_change_permission(self, request, obj=None):
        return False if obj and not obj.is_user_managed else True
            

class StructureAdminInline(admin.TabularInline):
    model = StructureService

    def has_add_permission(self, request):
        return True if STRUCTURES_DEVELOPER_MODE else False

    def has_change_permission(self, request, obj=None):
        return True if STRUCTURES_DEVELOPER_MODE else False

    def has_delete_permission(self, request, obj=None):
        return True if STRUCTURES_DEVELOPER_MODE else False


class OwnerCorporationsFilter(admin.SimpleListFilter):
    """Custom filter to filter on corporations from owners only"""
    title = 'owner corporation'
    parameter_name = 'owner_corporation_id__exact'

    def lookups(self, request, model_admin):
        qs = EveCorporationInfo.objects\
            .filter(owner__isnull=False)\
            .values('corporation_id', 'corporation_name')\
            .distinct()\
            .order_by(Lower('corporation_name'))
        return tuple(
            [(x['corporation_id'], x['corporation_name']) for x in qs]
        )

    def queryset(self, request, qs):
        if self.value() is None:
            return qs.all()
        else:
            return qs.filter(owner__corporation__corporation_id=self.value())


class OwnerAllianceFilter(admin.SimpleListFilter):
    """Custom filter to filter on alliances from owners only"""
    title = 'owner alliance'
    parameter_name = 'owner_alliance_id__exact'

    def lookups(self, request, model_admin):
        qs = EveAllianceInfo.objects\
            .filter(evecorporationinfo__owner__isnull=False)\
            .values('alliance_id', 'alliance_name')\
            .distinct()\
            .order_by(Lower('alliance_name'))
        return tuple(
            [(x['alliance_id'], x['alliance_name']) for x in qs]
        )

    def queryset(self, request, qs):
        if self.value() is None:
            return qs.all()
        else:        
            return qs\
                .filter(owner__corporation__alliance__alliance_id=self.value())
            

@admin.register(Structure)
class StructureAdmin(admin.ModelAdmin):
    show_full_result_count = True
    list_select_related = True
    search_fields = [
        'name',
        'owner__corporation__corporation_name',
        'eve_solar_system__name'
    ]
    ordering = ['name']
    list_display = (
        'name',
        '_owner',
        '_location',
        '_type',
        '_tags'
    )
    list_filter = (
        OwnerCorporationsFilter,
        OwnerAllianceFilter,
        ('eve_solar_system', admin.RelatedOnlyFieldListFilter),
        (
            'eve_solar_system__eve_constellation__eve_region',
            admin.RelatedOnlyFieldListFilter
        ),
        ('eve_type', admin.RelatedOnlyFieldListFilter),
        ('eve_type__eve_group', admin.RelatedOnlyFieldListFilter),
        (
            'eve_type__eve_group__eve_category',
            admin.RelatedOnlyFieldListFilter
        ),
        ('tags', admin.RelatedOnlyFieldListFilter),
    )

    actions = ('add_default_tags', 'remove_user_tags', 'update_generated_tags')

    def _owner(self, structure):
        return format_html(
            '{}<br>{}',
            structure.owner.corporation,
            structure.owner.corporation.alliance
        )

    def _location(self, structure):
        if structure.eve_moon:
            location_name = structure.eve_moon.name
        elif structure.eve_planet:
            location_name = structure.eve_planet.name
        else:
            location_name = structure.eve_solar_system.name
        return format_html(
            '{}<br>{}',
            location_name,
            structure.eve_solar_system.eve_constellation.eve_region
        )

    def _type(self, structure):
        return format_html(
            '{}<br>{}', structure.eve_type, structure.eve_type.eve_group
        )

    def _tags(self, structure):
        tag_names = [x.name for x in structure.tags.all()]
        if tag_names:
            return tag_names
        else:
            return None

    _tags.short_description = 'Tags'

    def has_add_permission(self, request):
        return True if STRUCTURES_DEVELOPER_MODE else False

    def add_default_tags(self, request, queryset):
        structure_count = 0
        tags = StructureTag.objects.filter(is_default=True)
        for structure in queryset:
            for tag in tags:
                structure.tags.add(tag)
            structure_count += 1

        self.message_user(
            request,
            'Added {:,} default tags to {:,} structures'.format(
                tags.count(),
                structure_count
            ))

    add_default_tags.short_description = \
        "Add default tags to selected structures"

    def remove_user_tags(self, request, queryset):
        structure_count = 0
        for structure in queryset:
            for tag in structure.tags.filter(is_user_managed=True):
                structure.tags.remove(tag)
            structure_count += 1

        self.message_user(
            request,
            'Removed all user tags from {:,} structures'.format(structure_count))

    remove_user_tags.short_description = \
        "Remove user tags for selected structures"

    def update_generated_tags(self, request, queryset):
        structure_count = 0
        for structure in queryset:
            structure.update_generated_tags(recreate_tags=True)
            structure_count += 1

        self.message_user(
            request,
            'Updated all generated tags for {:,} structures'.format(structure_count))

    update_generated_tags.short_description = \
        "Update generated tags for selected structures"

    if not STRUCTURES_DEVELOPER_MODE:
        readonly_fields = tuple([
            x.name for x in Structure._meta.get_fields()
            if isinstance(x, models.fields.Field)
            and x.name not in ['tags']
        ])

    fieldsets = (
        (None, {
            'fields': (
                'name', 'owner', 'eve_solar_system', 'eve_type', 'tags'
            )
        }),
        ('Status', {
            'classes': ('collapse',),
            'fields': (
                'state',
                ('state_timer_start', 'state_timer_end', ),
                'unanchors_at',
                'fuel_expires_at',
                'last_online_at',
            )
        }),
        ('Reinforcement', {
            'classes': ('collapse',),
            'fields': (
                ('reinforce_hour', ),
                (
                    'next_reinforce_hour',
                    'next_reinforce_weekday',
                    'next_reinforce_apply'
                ),
            )
        }),
        ('Position', {
            'classes': ('collapse',),
            'fields': ('position_x', 'position_y', 'position_z')
        }),
        (None, {
            'fields': (
                ('id', 'last_updated_at', )
            )
        }),
    )
    inlines = (StructureAdminInline, )

    def formfield_for_manytomany(self, db_field, request, **kwargs):
        """only show custom tags in dropdown"""
        if db_field.name == "tags":
            kwargs["queryset"] = StructureTag.objects.filter(is_user_managed=True)
        
        return super(StructureAdmin, self).formfield_for_manytomany(
            db_field, request, **kwargs
        )


@admin.register(Webhook)
class WebhookAdmin(admin.ModelAdmin):
    ordering = ['name']
    list_display = ('name', 'webhook_type', 'is_active', 'is_default')
    list_filter = ('webhook_type', 'is_active')
    save_as = True
    actions = (
        'test_notification',
        'activate',
        'deactivate'
    )

    def test_notification(self, request, queryset):
        for obj in queryset:
            tasks.send_test_notifications_to_webhook.delay(
                obj.pk,
                user_pk=request.user.pk
            )
            self.message_user(
                request,
                'Initiated sending test notification to webhook "{}". '
                'You will receive a report on completion.'.format(obj)
            )

    test_notification.short_description = \
        "Send test notification to selected webhooks"

    def activate(self, request, queryset):
        for obj in queryset:
            obj.is_active = True
            obj.save()

            self.message_user(
                request,
                'You have activated profile "{}"'.format(obj)
            )

    activate.short_description = 'Activate selected profiles'

    def deactivate(self, request, queryset):
        for obj in queryset:
            obj.is_active = False
            obj.save()

            self.message_user(
                request,
                'You have de-activated profile "{}"'.format(obj)
            )

    deactivate.short_description = 'Deactivate selected profiles'



# Generated by Django 2.2.12 on 2020-05-26 14:09

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('structures', '0010_sovereignty_map'),
    ]

    operations = [
        migrations.RenameField(
            model_name='structure',
            old_name='fuel_expires',
            new_name='fuel_expires_at',
        ),
        migrations.RenameField(
            model_name='structure',
            old_name='last_updated',
            new_name='last_updated_at',
        ),
        migrations.AddField(
            model_name='structure',
            name='created_at',
            field=models.DateTimeField(default=django.utils.timezone.now, help_text='date this structure was received from ESI for the first time'),
        ),
        migrations.AddField(
            model_name='structure',
            name='last_online_at',
            field=models.DateTimeField(blank=True, default=None, help_text="date this structure had any of it's services online", null=True),
        ),
    ]



from django.core.management.base import BaseCommand

from ...models import (
    EveCategory,
    EveGroup,
    EveType,
    EveRegion,
    EveConstellation,
    EveSolarSystem,
    EveMoon    
)


def get_input(text):
    """wrapped input to enable unit testing / patching"""
    return input(text)


class Command(BaseCommand):
    help = 'Updates Eve Online SDE data'
    
    def _update_models(self):
        """updates all SDE models from ESI and provides progress output"""
        models = [
            EveCategory,
            EveGroup,
            EveType,
            EveRegion,
            EveConstellation,
            EveSolarSystem,
            EveMoon,        
        ]
        model_count = 0
        for EveModel in models:
            total_objects = EveModel.objects.count()                        
            model_count += 1
            self.stdout.write(
                'Updating %d objects of %s (%d/%d)...' % (
                    total_objects, 
                    EveModel.__name__, 
                    model_count, 
                    len(models)
                )
            )            
            count_updated = EveModel.objects.update_all_esi()
            if count_updated < total_objects:
                self.stdout.write(self.style.DANGER(
                    'Only %d objects updated due to an error.' % count_updated
                ))
                
    def handle(self, *args, **options):
        self.stdout.write(
            'This command will reload all local EVE Online SDE data from '
            'the server. This process can take a while to complete.'            
        )
        user_input = get_input('Are you sure you want to proceed? (Y/n)?')
        if user_input == 'Y':
            self.stdout.write('Starting update. Please stand by.')
            self._update_models()
            self.stdout.write('Update completed!')
        else:
            self.stdout.write('Aborted')



import os 
from setuptools import find_packages, setup

from structures import __version__


# read the contents of your README file
this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='aa-structures',
    version=__version__,
    packages=find_packages(),
    include_package_data=True,
    license='MIT',
    description='App for managing Eve Online structures with Alliance Auth',
    long_description=long_description,
    long_description_content_type='text/markdown',    
    author='Erik Kalkoken',
    author_email='kalkoken87@gmail.com',
    url='https://gitlab.com/ErikKalkoken/aa-structures',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 2.2',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License', 
        'Operating System :: OS Independent',
        'Programming Language :: Python',        
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
    python_requires='~=3.5',
    install_requires=[ 
        'django>=2.2,<3.0',
        'django-esi<2.0',
        'dhooks-lite>=0.3.0',
        'django-navhelper',        
        'django-multiselectfield'
    ]
)



import urllib

from django.contrib.auth.decorators import login_required, permission_required
from django.db import transaction
from django.http import HttpResponse, JsonResponse, HttpResponseServerError
from django.shortcuts import render, redirect, reverse
from django.utils.html import format_html, mark_safe, escape
from django.utils.timezone import now
from django.utils.translation import gettext_lazy

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from allianceauth.eveonline.evelinks import dotlan
from allianceauth.services.hooks import get_extension_logger

from esi.decorators import token_required

from . import tasks, __title__
from .app_settings import (
    STRUCTURES_ADMIN_NOTIFICATIONS_ENABLED,
    STRUCTURES_SHOW_FUEL_EXPIRES_RELATIVE,
    STRUCTURES_DEFAULT_TAGS_FILTER_ENABLED
)
from .forms import TagsFilterForm
from .models import Owner, Structure, StructureTag, StructureService, Webhook
from .utils import (
    format_html_lazy, 
    messages_plus, 
    DATETIME_FORMAT, 
    notify_admins, 
    LoggerAddTag,
    timeuntil_str,
    add_no_wrap_html,
    yesno_str,
    create_bs_label_html
)


logger = LoggerAddTag(get_extension_logger(__name__), __title__)
STRUCTURE_LIST_ICON_RENDER_SIZE = 64
STRUCTURE_LIST_ICON_OUTPUT_SIZE = 32
QUERY_PARAM_TAGS = 'tags'


@login_required
@permission_required('structures.basic_access')
def index(request):
    url = reverse('structures:structure_list')
    if STRUCTURES_DEFAULT_TAGS_FILTER_ENABLED:
        params = {
            QUERY_PARAM_TAGS: ','.join([
                x.name
                for x in StructureTag.objects.filter(is_default=True)
            ])
        }
        url += '?{}'.format(urllib.parse.urlencode(params))
    return redirect(url)


@login_required
@permission_required('structures.basic_access')
def structure_list(request):
    """main view showing the structure list"""

    active_tags = list()
    if request.method == 'POST':
        form = TagsFilterForm(data=request.POST)
        if form.is_valid():
            for name, activated in form.cleaned_data.items():
                if activated:
                    active_tags.append(StructureTag.objects.get(name=name))

            url = reverse('structures:structure_list')
            if active_tags:
                params = {
                    QUERY_PARAM_TAGS: ','.join([x.name for x in active_tags])
                }
                url += '?{}'.format(urllib.parse.urlencode(params))
            return redirect(url)
    else:
        tags_raw = request.GET.get(QUERY_PARAM_TAGS)
        if tags_raw:
            tags_parsed = tags_raw.split(',')
            active_tags = [
                x for x in StructureTag.objects.all()
                if x.name in tags_parsed
            ]

        form = TagsFilterForm(initial={x.name: True for x in active_tags})

    context = {
        'page_title': gettext_lazy(__title__),
        'active_tags': active_tags,
        'tags_filter_form': form,
        'tags_exist': StructureTag.objects.exists()
    }
    return render(request, 'structures/structure_list.html', context)


class StructuresRowBuilder:
    """This class build the HTML table rows from structure objects"""
    def __init__(self):
        self._row = None
        self._structure = None
    
    def convert(self, structure) -> dict:
        self._row = {
            'structure_id': structure.id
        }
        self._structure = structure
        self._build_owner()
        self._build_location()
        self._build_type()
        self._build_name()
        self._build_services()
        self._build_reinforcement_infos()
        self._build_fuel_infos()
        self._build_online_infos()
        self._build_state()
        return self._row

    def _build_owner(self):
        corporation = self._structure.owner.corporation
        if corporation.alliance:
            alliance_name = corporation.alliance.alliance_name
        else:
            alliance_name = ""

        self._row['owner'] = format_html(
            '<a href="{}">{}</a><br>{}',
            dotlan.corporation_url(corporation.corporation_name),
            corporation.corporation_name,
            alliance_name
        )
        self._row['corporation_icon'] = format_html(
            '<img src="{}" width="{}" height="{}"/>',
            corporation.logo_url(size=STRUCTURE_LIST_ICON_RENDER_SIZE),
            STRUCTURE_LIST_ICON_OUTPUT_SIZE,
            STRUCTURE_LIST_ICON_OUTPUT_SIZE,
        )
        self._row['alliance_name'] = alliance_name
        self._row['corporation_name'] = corporation.corporation_name
        
    def _build_location(self):        
        solar_system = self._structure.eve_solar_system
        
        # location
        self._row['region_name'] = \
            solar_system.eve_constellation\
            .eve_region.name_localized
        self._row['solar_system_name'] = \
            solar_system.name_localized
        solar_system_url = dotlan.solar_system_url(
            solar_system.name
        )
        if self._structure.eve_moon:
            location_name = self._structure.eve_moon.name_localized
        elif self._structure.eve_planet:
            location_name = self._structure.eve_planet.name_localized
        else:
            location_name = self._row['solar_system_name']

        self._row['location'] = format_html(
            '<a href="{}">{}</a><br>{}',
            solar_system_url,
            add_no_wrap_html(location_name),
            add_no_wrap_html(self._row['region_name'])
        )

    def _build_type(self):
        structure_type = self._structure.eve_type
        # category
        my_group = structure_type.eve_group
        self._row['group_name'] = my_group.name_localized
        if my_group.eve_category:
            my_category = my_group.eve_category
            self._row['category_name'] = my_category.name_localized
            self._row['is_starbase'] = my_category.is_starbase
        else:
            self._row['category_name'] = ''
            self._row['is_starbase'] = None

        # type icon
        self._row['type_icon'] = format_html(
            '<img src="{}" width="{}" height="{}"/>',
            structure_type.icon_url(
                size=STRUCTURE_LIST_ICON_RENDER_SIZE
            ),
            STRUCTURE_LIST_ICON_OUTPUT_SIZE,
            STRUCTURE_LIST_ICON_OUTPUT_SIZE,
        )

        # type name
        self._row['type_name'] = structure_type.name_localized
        self._row['type'] = format_html(
            '{}<br>{}',
            add_no_wrap_html(self._row['type_name']),
            add_no_wrap_html(self._row['group_name'])
        )

        # poco
        self._row['is_poco'] = structure_type.is_poco

    def _build_name(self):
        self._row['structure_name'] = escape(self._structure.name)                
        tags = []        
        if self._structure.tags:
            tags += [
                x.html for x in self._structure.tags.all()
            ]
            self._row['structure_name'] += format_html(
                '<br>{}', mark_safe(' '.join(tags))
            )

    def _build_services(self):
        if self._row['is_poco'] or self._row['is_starbase']:
            self._row['services'] = gettext_lazy('N/A')
        else:
            services = list()
            services_qs = \
                self._structure.structureservice_set.all().order_by('name')
            for service in services_qs:
                service_name = add_no_wrap_html(format_html(
                    '<small>{}</small>', service.name_localized
                ))
                if service.state == StructureService.STATE_OFFLINE:
                    service_name = format_html('<del>{}</del>', service_name)
                
                services.append(service_name)
            self._row['services'] = '<br>'.join(services)

    def _build_reinforcement_infos(self):
        self._row['is_reinforced'] = self._structure.is_reinforced
        self._row['is_reinforced_str'] = \
            yesno_str(self._structure.is_reinforced)

        if self._row['is_starbase']:
            self._row['reinforcement'] = gettext_lazy('N/A')
        else:
            if self._structure.reinforce_hour:
                self._row['reinforcement'] = '{:02d}:00'.format(
                    self._structure.reinforce_hour
                )
            else:
                self._row['reinforcement'] = ''

    def _build_fuel_infos(self):
        if self._structure.eve_type.is_poco:
            fuel_expires_display = 'N/A'
            fuel_expires_timestamp = None
        elif self._structure.is_low_power:
            fuel_expires_display = format_html_lazy(
                create_bs_label_html(
                    self._structure.get_power_mode_display(), 'warning'
                )
            )                    
            fuel_expires_timestamp = None
        elif self._structure.is_abandoned:
            fuel_expires_display = format_html_lazy(
                create_bs_label_html(
                    self._structure.get_power_mode_display(), 'danger'
                )
            )
            fuel_expires_timestamp = None
        elif self._structure.is_maybe_abandoned:
            fuel_expires_display = format_html_lazy(
                create_bs_label_html(
                    self._structure.get_power_mode_display(), 'warning'
                )
            )                    
            fuel_expires_timestamp = None
        elif self._structure.fuel_expires_at:
            fuel_expires_timestamp = self._structure.fuel_expires_at.isoformat()
            if STRUCTURES_SHOW_FUEL_EXPIRES_RELATIVE:
                fuel_expires_display = timeuntil_str(
                    self._structure.fuel_expires_at - now()
                )
                if not fuel_expires_display:
                    fuel_expires_display = '?'
                    fuel_expires_timestamp = None
            else:
                if self._structure.fuel_expires_at >= now():
                    fuel_expires_display = \
                        self._structure.fuel_expires_at.strftime(DATETIME_FORMAT)
                else:
                    fuel_expires_display = '?'
                    fuel_expires_timestamp = None                
        else:
            fuel_expires_display = gettext_lazy('N/A')
            fuel_expires_timestamp = None

        self._row['fuel_expires_at'] = {
            'display': add_no_wrap_html(fuel_expires_display),
            'timestamp': fuel_expires_timestamp
        }

    def _build_online_infos(self):        
        self._row['power_mode_str'] = self._structure.get_power_mode_display()
        if self._structure.eve_type.is_poco:
            last_online_at_display = 'N/A'
            last_online_at_timestamp = None
        elif self._structure.is_full_power:
            last_online_at_display = format_html_lazy(
                create_bs_label_html(
                    self._structure.get_power_mode_display(), 'success'
                )
            )
            last_online_at_timestamp = None
        elif self._structure.is_maybe_abandoned:
            last_online_at_display = format_html_lazy(
                create_bs_label_html(
                    self._structure.get_power_mode_display(), 'warning'
                )
            )                    
            last_online_at_timestamp = None
        elif self._structure.is_abandoned:
            last_online_at_display = format_html_lazy(
                create_bs_label_html(
                    self._structure.get_power_mode_display(), 'danger'
                )
            )
            last_online_at_timestamp = None
        elif self._structure.last_online_at:
            last_online_at_timestamp = \
                self._structure.last_online_at.isoformat()
            if STRUCTURES_SHOW_FUEL_EXPIRES_RELATIVE:
                last_online_at_display = timeuntil_str(
                    now() - self._structure.last_online_at
                )
                if not last_online_at_display:
                    last_online_at_display = '?'
                    last_online_at_timestamp = None
                else:
                    last_online_at_display = '- ' + last_online_at_display
            else:
                last_online_at_display = \
                    self._structure.last_online_at.strftime(DATETIME_FORMAT)
        else:
            last_online_at_display = '-'
            last_online_at_timestamp = None

        self._row['last_online_at'] = {
            'display': add_no_wrap_html(last_online_at_display),
            'timestamp': last_online_at_timestamp
        }

    def _build_state(self):
        self._row['state_str'] = self._structure.get_state_display()
        self._row['state_details'] = self._row['state_str']
        if self._structure.state_timer_end:
            self._row['state_details'] += format_html(
                '<br>{}',
                add_no_wrap_html(
                    self._structure.state_timer_end.strftime(DATETIME_FORMAT)
                )
            )


@login_required
@permission_required('structures.basic_access')
def structure_list_data(request):
    """returns structure list in JSON for AJAX call in structure_list view"""

    structure_rows = list()
    row_converter = StructuresRowBuilder()
    for structure in _structures_query_for_user(request):        
        structure_rows.append(row_converter.convert(structure))

    return JsonResponse(structure_rows, safe=False)


def _structures_query_for_user(request):
    """returns query according to users permissions and current tags"""
    tags_raw = request.GET.get(QUERY_PARAM_TAGS)
    if tags_raw:
        tags = tags_raw.split(',')
    else:
        tags = None

    if request.user.has_perm('structures.view_all_structures'):
        structures_query = Structure.objects.all().select_related()
        if tags:
            structures_query = structures_query\
                .filter(tags__name__in=tags)\
                .distinct()

    else:
        corporation_ids = {
            character.character.corporation_id
            for character in request.user.character_ownerships.all()
        }
        corporations = list(
            EveCorporationInfo.objects.filter(
                corporation_id__in=corporation_ids
            )
        )
        if request.user.has_perm('structures.view_alliance_structures'):
            alliances = {
                corporation.alliance
                for corporation in corporations if corporation.alliance
            }
            for alliance in alliances:
                corporations += alliance.evecorporationinfo_set.all()

            corporations = list(set(corporations))

        structures_query = Structure.objects\
            .filter(owner__corporation__in=corporations) \
            .select_related()

    return structures_query


@login_required
@permission_required('structures.add_structure_owner')
@token_required(scopes=Owner.get_esi_scopes())
def add_structure_owner(request, token):
    token_char = EveCharacter.objects.get(character_id=token.character_id)

    success = True
    try:
        owned_char = CharacterOwnership.objects.get(
            user=request.user,
            character=token_char
        )
    except CharacterOwnership.DoesNotExist:
        messages_plus.error(
            request,            
            format_html(
                gettext_lazy(
                    'You can only use your main or alt characters '
                    'to add corporations. '
                    'However, character %s is neither. '
                ) % format_html(
                    '<strong>{}</strong>', token_char.character_name
                )
            )
        )
        success = False

    if success:
        try:
            corporation = EveCorporationInfo.objects.get(
                corporation_id=token_char.corporation_id
            )
        except EveCorporationInfo.DoesNotExist:
            corporation = EveCorporationInfo.objects.create_corporation(
                token_char.corporation_id
            )

        with transaction.atomic():
            owner, _ = Owner.objects.update_or_create(
                corporation=corporation,
                defaults={
                    'character': owned_char
                }
            )
            default_webhooks = Webhook.objects.filter(is_default=True)
            if default_webhooks:
                for webhook in default_webhooks:
                    owner.webhooks.add(webhook)
                owner.save()

        tasks.update_structures_for_owner.delay(
            owner_pk=owner.pk, user_pk=request.user.pk
        )
        messages_plus.info(
            request,            
            format_html(                
                gettext_lazy(
                    '%(corporation)s has been added with %(character)s '
                    'as sync character. We have started fetching structures '
                    'for this corporation. You will receive a report once '
                    'the process is finished.'
                ) % {
                    'corporation': format_html(
                        '<strong>{}</strong>', owner
                    ),
                    'character': format_html(
                        '<strong>{}</strong>', 
                        owner.character.character.character_name
                    )
                }
            )
        )
        if STRUCTURES_ADMIN_NOTIFICATIONS_ENABLED:
            notify_admins(                
                message=gettext_lazy(
                    '%(corporation)s was added as new '
                    'structure owner by %(user)s.') % {                    
                        'corporation': owner.corporation.corporation_name,
                        'user': request.user.username
                },
                title='{}: Structure owner added: {}'.format(
                    __title__,
                    owner.corporation.corporation_name
                )
            )
    return redirect('structures:index')


def service_status(request):
    """public view to 3rd party monitoring

    This is view allows running a 3rd party monitoring on the status
    of this services. Service will be reported as down if any of the
    configured structure or notifications syncs fails or is delayed
    """
    status_ok = True
    for owner in Owner.objects.filter(is_included_in_service_status=True):
        status_ok = status_ok and owner.are_all_syncs_ok

    if status_ok:
        return HttpResponse(gettext_lazy('service is up'))
    else:
        return HttpResponseServerError(gettext_lazy('service is down'))



# Generated by Django 2.2.5 on 2019-12-03 15:25

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('structures', '0004_auto_20191129_0308'),
    ]

    operations = [
        migrations.AddField(
            model_name='notification',
            name='created',
            field=models.DateTimeField(blank=True, default=None, help_text='Date when this notification was first received from ESI', null=True),
        ),
    ]
